/*global jQuery, sas, PfdXmlParser, $, go*/
//var jQuery, sas, PfdXmlParser, $;
//jQuery.sap.declare('sas.risk.irm.pfd');
//sas.risk.irm.pfd = {};
//jQuery.sap.require('sas.risk.irm.SIIUtils');
jQuery.sap.declare('sas.risk.irm.pfd.PfdXmlParser');

sas.risk.irm.pfd.PfdXmlParser = (function () {
    "use strict";
    var BPMN_DI_URL = 'http://www.omg.org/spec/BPMN/20100524/DI',
        BPMN_MODEL_URL = 'http://www.omg.org/spec/BPMN/20100524/MODEL',
        DATA_INPUT_ASSOCIATION_TAG = 'dataInputAssociation',
        DATA_OUTPUT_ASSOCIATION_TAG = 'dataOutputAssociation',
        DATA_OBJECT_REFERENCE_TAG = 'dataObjectReference',
        DATA_STATE_TAG = 'dataState',
        DEFINITIONS_TAG = 'definitions',
        ATTRIBUTE_KEY_DATA_OBJECT_REF = 'dataObjectRef',
        ATTRIBUTE_KEY_TARGET_REF = 'targetRef',
        ATTRIBUTE_KEY_SOURCE_REF = 'sourceRef',
        ATTRIBUTE_KEY_NAME = 'name',
        ATTRIBUTE_KEY_ID = 'id',
        ATTRIBUTE_KEY_ISCOLLECTION = 'isCollection',
        ATTRIBUTE_KEY_BPMN_ELEMENT = 'bpmnElement',
        ATTRIBUTE_VALUE_PARTITION = 'partition',
        PROPERTY_KEY_TAGNAME = 'tagName',
        BPMNSHAPE_TAGNAME = 'BPMNShape',
        BPMNEDGE_TAGNAME = 'BPMNEdge',
        SELECTOR_DOCUMENTATION = 'documentation',
        SELECTOR_PROPERTY = 'property',
        PORT_VALUE_DATA_OUTPUT = 'dataOutput',
        PORT_VALUE_DATA_INPUT = 'dataInput',
        PORT_VALUE_FLOW_OUTPUT = 'flowOutput',
        PORT_VALUE_FLOW_INPUT = 'flowInput',
        bpmnDiNamespace = "bpmndi",
        bpmnModelNamespace = "";

    /**
     * This method walks over each of the namespace declarations in the definitions element
     * (each one starts with xmlns:) and finds the one that matches the DI and MODEL schema.  Once that is found
     * the localName will hold the part of the name space after the prefix (xmlns:). This will be the real
     * namespace for the di/model elements respectively.
     * @param docRoot The document to be searched
     */
    function resolveNameSpaces(docRoot) {
        var definitions =  docRoot.getElementsByTagNameNS ? docRoot.getElementsByTagNameNS("*",DEFINITIONS_TAG)[0] : docRoot.getElementsByTagName(DEFINITIONS_TAG)[0],
            attributes = definitions.attributes,
            i,
            attribute;
        for (i = 0; i < attributes.length; i++) {
            attribute = attributes[i];
            if (attribute.value === BPMN_DI_URL) {
                bpmnDiNamespace = attribute.localName;
            }
            if (attribute.value === BPMN_MODEL_URL) {
                bpmnModelNamespace = attribute.localName;
                if (bpmnModelNamespace == undefined || bpmnModelNamespace === "xmlns") {
                    bpmnModelNamespace = "";
                }
            }
        }
    }

    /**
     * Our parser has to reference elements in the process section of the BMPN file, but different tools use
     * different namespaces for the elements in that area.  So this method strips off the namespace after we have
     * found an element so that the rest of the code works.  If the element doesn't use a namespace then the passed
     * in value is just returned.
     * @param value A element tag whose value may have a namespace in front of it.
     * @returns {*} The tag of the element without the name space attached.
     */
    function removeNameSpace(value) {
        var index = value.lastIndexOf(':');
        if (index !== -1) {
            return value.substring(index+1);
        }
        return value;
    }

    /**
     * To find an element by tag using jQuery we have to create the entire tag name
     * If a namespace is used then we have escape the namespace separator so jQuery
     * doesn't use that as it's selector.  This method creates that tag name and
     * properly appends the namespace if one was found for the MODEL namespace while
     * also escaping the namespace separator correctly.
     * @param tag
     * @returns {*}
     */
    function getTagWithNameSpaceForJQuery(tag){
        if (bpmnModelNamespace && bpmnModelNamespace.length > 0) {
            return bpmnModelNamespace+'\\:'+tag;
        }
        return tag;
    }

    /**
     * Takes an id string and escapes special characters that are allowed.  Right now
     * this is periods ('.'), pluses ('+'), percents ('%'), and asterisks ('*')
     * @param {string} id - The id from the bpmn file
     * @returns {XML|*|string|void} - The id with ".", "+", "%", and "*" escaped
     */
    function sanitizeId(id) {
        // Want to replace any special characters with double backspaces
        var newString = id.replace(/\./g,"\\.");
        newString = newString.replace(/\+/g,"\\+");
        newString = newString.replace(/\*/g,"\\*");
        newString = newString.replace(/\%/g,"\\%");
        return newString;
    }

    /**
     * Each browser's DOM XML parser works slightly different than the others, this method
     * abstracts out the code to get an element by it's tag name so we don't have to
     * do a bunch of checking for different browser implementations.
     * @param {Document} documentRoot - The xml document to search in
     * @param {string} tagPrefix - The prefix of the tag name
     * @param {string} tagSuffix - The suffix of the tag name
     * @returns {NodeList|*} - The list of matching elements
     */
    function getElementsByTagNameAgnostic(documentRoot, tagPrefix, tagSuffix) {
        var tagName = tagPrefix + ":" + tagSuffix,
            elements;
        if (tagPrefix === null) {
            tagName = tagSuffix;
        }
        elements = documentRoot.getElementsByTagName(tagName);
        if (!elements || elements === null || elements.length === 0) {
            elements = documentRoot.getElementsByTagName(tagSuffix);
        }
//        if (!elements || elements === null) {
//            elements = documentRoot.getElementsByTagName(BPMN_URL, tagSuffix);
//        }
        return elements;
    }

    /**
     * Retrieves the element with an id attribute having the passed in nodeId for
     * the passed in document.  The id is first sanitized for special characters
     * @param {string} nodeId The id of the node to search for
     * @param {Document} docRoot The xml document to be searched in
     * @returns {NodeList|*} - The matching node
     */
    function getNodeWithId(nodeId, docRoot) {
        var sanitizedId = sanitizeId(nodeId);
        return $('#' + sanitizedId, docRoot);
    }

    /**
     * Returns the name for the element.  This is the internal name, it is not the
     * name for displaying.
     * @param {jQuery} element The element to get the name from
     * @param {string} nameAttributeKey The key for the name attribute
     * @param {boolean} includeIds Should the id string be appended to the name
     * @returns {string} The nodes name
     */
    function getElementName(element, nameAttributeKey, includeIds) {
        var att = element.attr(nameAttributeKey),
            name = "",
            re;
        if (att) {
            name = att;
            re = new RegExp(' ', 'g');
            name = name.replace(re, '\n');
        }
        if (includeIds) {
            name = name + element.attr(ATTRIBUTE_KEY_ID);
        }
        return name;
    }

    /**
     * Returns the display name for the element.  This is the name that should be
     * shown on the flow diagram
     * @param {jQuery} element The element to get the display name from
     * @param {string} nameAttributeKey The key for the display name
     * @param {boolean} Does the element represent a data node
     * @param  {Object} messagesList Map of nameKey to displayable names
     * @param {boolean} shouldStripExtension To look the displayable name up should we strip the extension first? Pass
     * true when you are dealing with a data node, false otherwise.
     * @returns {string} A displayable name for the element
     */
    function getElementDisplayName(element, nameAttributeKey, isDataNode, messagesList, shouldStripExtension) {
        var att = element.attr(nameAttributeKey),
            name = "",
            re;
        if (att) {
            name = att;
            //re = new RegExp(' ', 'g');
            //name = name.replace(re, '\n');
        }
        if (shouldStripExtension) {
            name = SIIUtils.stripExtension(name);
        }
        return SIIUtils.getDisplayName(name, isDataNode, messagesList);
    }

    function getElementRawName(element, nameAttributeKey) {
        var att = element.attr(nameAttributeKey),
            name = "",
            re;
        if (att) {
            return att;
        }
        return null;
    }

    /**
     * Returns the datatype for the element with the given name attribute key.
     * At this time the result will either be "excel", "xbrl", "sasdataset" or ""
     * @param {jQuery} element The element to get the data type from
     * @param {string} nameAttributeKey The key for the datatype
     * @returns {string} The data type string for the element
     */
    function getDataType(element,nameAttributeKey){
        var name;
//            extension;
        var collectionFlag = element.attr(ATTRIBUTE_KEY_ISCOLLECTION);
        if (collectionFlag && collectionFlag === "true") {
            return SIIConstants.FILE_TYPE_COLLECTION;
        }
        name = getElementName(element, nameAttributeKey, false);

        // Determine if this data node represents a partition table
        if (isPartitioned(element)) {
            return SIIConstants.FILE_TYPE_PARTITION_TABLE;
        }
        return SIIUtils.getDataType(name);
    }

//    /**
//     * Determines whether the node passed in is empty or not.
//     * @param {Node} node The node to be checked
//     * @returns {boolean} true if the node is empty
//     */
//    function isEmptyNode(node) {
//        return /^\s*$/.test(node.nodeValue);
//    }

    /**
     * Returns the id of the task node that contains the passed in dataRefId (Note: I use this term for both
     * sourceRef and targetRef elements) in the passed in docRoot.
     * Note: The taskId is found as follows given the following xml fragment:
     * <serviceTask id="aaa" ...>
     *     <ioSpecification>
     *         <dataInput id="bbb"/>
     *         <dataOutput id="ccc"/>
     *     </ioSpecification>
     * </serviceTask>
     *
     * Assuming "ccc" is passed in as the dataRefId, we first find that element (which will be a dataOutput
     * element in the example).  Next we go to the parent which is an ioSpecification element. Finally we
     * get it's parent (which may be a serviceTask of subProcess element) and we return the id on that node
     * @param {string} dataRefId The id of the dataRef to find the associated taskNode id for
     * @param {Document} docRoot The document to be searched
     * @returns {*|jQuery} The id for the task node with the given dataRefId
     */
    function findTaskIdForDataRef(dataRefId, docRoot) {
        var element,
            ioSpecElement,
            taskElement,
            taskParentId;

        element = getNodeWithId(dataRefId, docRoot);
        if (element !== null) {
            // element is expected to be either a dataOutput or dataInput
            SIIUtils.log("Tag name for element: " + element.prop(PROPERTY_KEY_TAGNAME));

            // Get it's parent which will be a <ioSpecification>
            ioSpecElement = element.parent();
            SIIUtils.log("Tag name for immediateParent: " + ioSpecElement.prop(PROPERTY_KEY_TAGNAME));

            // Get it's parent which will be a <serviceTask> or a <subProcess>
            taskElement = ioSpecElement.parent();
            SIIUtils.log("Tag name for taskParent: " + taskElement.prop(PROPERTY_KEY_TAGNAME));

            // Get and return the id of the task/subProcess
            taskParentId = taskElement.attr(ATTRIBUTE_KEY_ID);
            return taskParentId;
        }
    }

    /**
     * Parses out the links (if any) for help documentation for a given node
     * @param element The element to be searched
     * @param faRootURL The url to the root of the federated area
     * @returns {{}} An object containing two elements:
     *   helpDocumentLink: for the absolute link to help documentation (will be brought up in a separate tab)
     *   helpTooltipLink: for the absolute link to tooltip documentation (essentially a page of html that will be shown in the tooltip)
     *   Note: If there is no help links then an empty object will be returned.
     */
    function getDocumentationElements(element, faRootURL) {
        var documentationElements = $(element).children(SELECTOR_DOCUMENTATION), i, docElement, returnObject, text;
        returnObject = {};
        for (i = 0; i < documentationElements.length; i++) {
            docElement = $(documentationElements[i]);
            text = docElement[0].textContent;
            if (text) {
                if ((/^DOCUMENT/).test(text.toUpperCase())){
                    // strip off the 'DOCUMENT=' part
                    returnObject.helpDocumentLink = text.substring(9);
                }
                else if ((/^TOOLTIP/).test(text.toUpperCase())) {
                    // strip off the 'TOOLTIP=' part
                    returnObject.helpTooltipLink = text.substring(8);
                }
            }
        }
        return returnObject;
    }

    /**
     * Searches the given element (which is assumed to be a non data node) for
     * its set of properties then finds and finds the one that starts with the
     * given prefix.  Returns the property value, minus the prefix.
     * @param {Node} element The element to be searched
     * @param {string} prefix of the property to be returned
     * @returns {string} The value of the property without the prefix
     */
    function getElementProperty(element, prefix) {
        var propertyElements = $(element).children(SELECTOR_PROPERTY), i, propertyName;

        for (i = 0; i < propertyElements.length; i++) {
            propertyName = $(propertyElements[i]).attr(ATTRIBUTE_KEY_NAME);

            var preamble = propertyName.split('=')[0];// take everything before the first =
            preamble = preamble.replace(/\s/g, ''); // remove all white space
            if (preamble.toUpperCase() === prefix.toUpperCase()) {
                var value = propertyName.split('=')[1]; // get everything after the =
                value = value.trim(); // remove whitespace before and after
                return value;
            }
            //if (propertyName.indexOf(prefix) === 0) {
            //    return propertyName.substring(prefix.length);
            //}
        }
        return null;
    }

    /**
     * Returns the id of the node that the passed in element flows to. In other words the diagram should
     * draw an arrow from the passed in element to the node with the BPMNElementId this method returns.
     * @param element The element to get the id of the node to go to
     * @param docRoot
     * @returns {*} The id of the target node or undefined if there is none.
     */
    function getTargetRef(element, docRoot) {
        var edgeType = element.prop(PROPERTY_KEY_TAGNAME),
            toId = element.attr(ATTRIBUTE_KEY_TARGET_REF), // Should we set this to undefined?
            sourceRef;
        edgeType = removeNameSpace(edgeType);
        if (edgeType === DATA_INPUT_ASSOCIATION_TAG || edgeType === DATA_OUTPUT_ASSOCIATION_TAG) {
            sourceRef = $(element).children(getTagWithNameSpaceForJQuery(ATTRIBUTE_KEY_TARGET_REF))[0];
            toId = $(sourceRef).text();
            if (edgeType === DATA_INPUT_ASSOCIATION_TAG) {
                // toId is an id to a dataInputRef for a shape so we have to find the task
                // that has this targetRef in it's list of dataInputRefs in it's inputSet
                // Note for now we find the id of the task that has a child element that looks like
                // <dataInput id="toId" isCollection="false"/> in its ioSpecification element
                toId = findTaskIdForDataRef(toId, docRoot);
            }
        }
        return toId;
    }

    /**
     * Returns the id of the node that the passed in element flows from. In other words the diagram should
     * draw an arrow to the passed in element from the node with the BPMNElementId this method returns.
     * @param element The element to get the id of the node from where it came
     * @param docRoot
     * @returns {*} The id of the source node or undefined if there is none.
     */
    function getSourceRef(element, docRoot) {
        var edgeType = element.prop(PROPERTY_KEY_TAGNAME),
            fromId = element.attr(ATTRIBUTE_KEY_SOURCE_REF), // Should we set this to undefined?
            sourceRef;
        edgeType = removeNameSpace(edgeType);
        if (edgeType === DATA_INPUT_ASSOCIATION_TAG || edgeType === DATA_OUTPUT_ASSOCIATION_TAG) {
            sourceRef = $(element).children(getTagWithNameSpaceForJQuery(ATTRIBUTE_KEY_SOURCE_REF))[0];
            fromId = $(sourceRef).text();
            if (edgeType === DATA_OUTPUT_ASSOCIATION_TAG) {
                // fromId is an id to a dataOutputRef for a shape so we have to find the task
                // that has this sourceRef in it's list of dataOutputRefs in it's outputSet
                // Note for now we find the id of the task that has a child element that looks like
                // <dataOutput id="fromId" isCollection="false"/> in its ioSpecification element
                SIIUtils.log("Searching for outputRef for id:"+fromId);
                fromId = findTaskIdForDataRef(fromId, docRoot);
            }
        }
        return fromId;
    }

    /**
     * Does the passed in element represent a partitioned table.  An element represents a partitioned table iff
     * it contains a child node that is a dataState tag and has a name attribute of 'partition'
     * @param element
     * @returns {boolean}
     */
    function isPartitioned(element) {
        var partitionReferences,
            partitionRef,
            nameValue;

        partitionReferences = $(element).children(getTagWithNameSpaceForJQuery(DATA_STATE_TAG));
        if (partitionReferences && partitionReferences.length > 0) {
            partitionRef = partitionReferences[0];
            nameValue = partitionRef.getAttribute(ATTRIBUTE_KEY_NAME);
            return (nameValue && nameValue.indexOf(ATTRIBUTE_VALUE_PARTITION) != -1);
        }
        return false;
    }

    /**
     * Isolates the code to check if the nodeType passed in is one of the ones we can render
     * If not, callers should set the type to be SIIConstants.NODE_UNKNOWN
     * @param nodeType The node type to check
     * @returns {boolean} true if the type is valid false otherwise
     */
    function isValidNodeType(nodeType){
        nodeType = removeNameSpace(nodeType);
        return (nodeType === SIIConstants.NODE_END ||
            nodeType === SIIConstants.NODE_SERVICE ||
            nodeType === SIIConstants.NODE_START ||
            nodeType === SIIConstants.FILE_TYPE_EXCEL ||
            nodeType === SIIConstants.FILE_TYPE_SASDATASET ||
            nodeType === SIIConstants.FILE_TYPE_XBRL ||
            nodeType === SIIConstants.NODE_SUBFLOW ||
            nodeType === SIIConstants.FILE_TYPE_COLLECTION ||
            nodeType === SIIConstants.FILE_TYPE_PARTITION_TABLE ||
            nodeType === SIIConstants.FILE_TYPE_XML ||
            nodeType === SIIConstants.FILE_TYPE_PDF ||
            nodeType === SIIConstants.FILE_TYPE_CSV ||
            nodeType === SIIConstants.FILE_TYPE_UNKNOWN
        );
    }

    // Return the public methods
    return {
        /**
         * Parses the passed in xml for node objects and builds a list of these as json
         * objects which are consumable by the PfdGoJSRenderer.  The results are returned
         * as an array and are also placed in the nodes array of this class.
         * @param {Document} xml The xml document to be parsed for nodes
         * @param {Object} messagesList Map containing bmpn name to localized name for display mappings
         * @param {string} faRootURL The url to the Federated Area's root
         * @returns {Array} A JSON array of the nodes for the xml document
         */
        parseNodes: function (xml, messagesList, faRootURL) {
            var i,
                shapes,
                shape,
                bpmnElementId,
                bounds,
                x,
                y,
                height,
                width,
                element,
                type,
                dataType,
                isPartitionTable = false,
                nodes,
//                groups,
//                from,
//                to,
                nameAttributeKey,
                iconType,
                nodeObject,
                displayName,
                docObject,
                pathInfo,
                sourceInfo,
                rawName,
                faName;
            // Make sure we have the correct namespace for the BPMNDiagram section
            resolveNameSpaces(xml); // Do this here for unit tests

            nodes = [];
//            groups = [];
            shapes = getElementsByTagNameAgnostic(xml, bpmnDiNamespace, BPMNSHAPE_TAGNAME);


            // The algorithm
            // Retrieve a list of BPMNShape objects
            // For each shape:
            //   1) Get the geometry information
            //   2) Retrieve the element with the id corresponding to the BPMNShape's "bpmnElement" attribute from the xml file
            //      The element will be either a serviceTask or dataObjectReference
            //      a: If the element is a serviceTask get the name, use the name to determine what type of task it is
            //      b: If the element is a dataObjectReference find the dataObject element that has an id which corresponds
            //         to the dataObjectReference's dataObjectRef attribute.  The dataObject element is used to determine
            //         if the dataNode will be a folder or reference a file and if so what type of file it is based on it's
            //         name's extension.
            //  3) Add the node attributes into the node array
            //

            for (i = 0; i < shapes.length; i++) {
                shape = shapes.item(i);
                docObject = null;

                // Get the BPM NElementId from the BPMNShape object
                bpmnElementId = shape.getAttribute(ATTRIBUTE_KEY_BPMN_ELEMENT);
                bounds = shape.childNodes[1]; //CHEAT: We are cheating here, as I know the second node is the one I want for the test xml !!!!
                x = bounds.getAttribute('x');
                var decimalLoc = x.indexOf(".");
                if (decimalLoc === -1) {
                    x = parseInt(x, 10);
                }
                else {
                    x = parseInt(x.substring(0, x.indexOf(".") + 3), 10);
                }

                y = bounds.getAttribute('y');
                var decimalLoc = y.indexOf(".");
                if (decimalLoc === -1) {
                    y = parseInt(y, 10);
                }
                else {
                    y = parseInt(y.substring(0, y.indexOf(".") + 3), 10);
                }

                height = parseInt(bounds.getAttribute('height'), 10);
                width = parseInt(bounds.getAttribute('width'), 10);

                // Find element with the BPMNElementId we got above
                element = getNodeWithId(bpmnElementId, xml);

                // Get the tag name off that element.  We expect this to be either
                // dataObjectReference or serviceTask.  This tells us what type of
                // node we are building
                type = element.prop(PROPERTY_KEY_TAGNAME);
                type = removeNameSpace(type);

                dataType = "";
//                var nodeGroup = element.attr('group');

                // NOTE: A lot of this back-tracing crap could be eliminated if we didn't have to use the
                // name to determine the type of the node.  Since the metadata includes the BPMNElementId we could
                // lazily determine names and display names when the nodes are being plotted in the renderer
                // itself, but for now we have to do it the hard way.
                // We don't want "lane" objects
                if (type !== "lane") {
                    nameAttributeKey = ATTRIBUTE_KEY_NAME;
                    iconType = getElementProperty(element, "type");

                    // If the tagname for the element is a dataObjectReference then we are
                    // dealing with a data node
                    if (type === DATA_OBJECT_REFERENCE_TAG) {
                        SIIUtils.log("Element: "+element);
                        // Retrieve the dataObjectRef attribute
                        var dataObjectRefId = getElementName(element, ATTRIBUTE_KEY_DATA_OBJECT_REF, false);

                        // Determine if this data node represents a partition table
                        isPartitionTable = isPartitioned(element);

                        // Use the dataObjectRef attribute as an id to find the corresponding dataObject
                        // element which will hold the data node's real name.
                        element = getNodeWithId(dataObjectRefId, xml); // IMPORTANT HERE!!! We are INTENTIONALLY changing the object that element points to!!

                        // Now we can get the type of the data object based on it's filename and collection state
                        SIIUtils.xmlToString(element);
                        displayName = getElementDisplayName(element,nameAttributeKey,true,messagesList,true);

                        // Determine the data node's type
                        if (isPartitionTable) {
                            type = SIIConstants.FILE_TYPE_PARTITION_TABLE;
                        }
                        else {
                            type = getDataType(element,nameAttributeKey);
                        }
                    }
                    else {
                        displayName = getElementDisplayName(element,nameAttributeKey,false,messagesList,false);
                        rawName = getElementRawName(element, nameAttributeKey);
                        // Get document tags
                        docObject = getDocumentationElements(element, faRootURL);
                        pathInfo = getElementProperty(element, "path");
                        sourceInfo = getElementProperty(element, "source");
                    }

                    // If the type is not one of the node types we know about we need to make sure we
                    // "self-heal" the node type so when we later plot it we can.
                    if (! isValidNodeType(type)) {
                        type = SIIConstants.NODE_UNKNOWN;
                    }
                    var name = getElementName(element, nameAttributeKey, false);
                    nodeObject = {
                        type: type,
                        dataType: dataType,
                        id: bpmnElementId,
                        bounds: {
                            x: x,
                            y: y,
                            width: width,
                            height: height
                        },
                        name: name,
                        displayName: displayName
                    };
                    if (iconType) {
                        nodeObject.iconType = iconType;
                    }

                    if (docObject) {
                        nodeObject.helpLinks = docObject;
                    }
                    if (pathInfo) {
                        nodeObject.pathInfo = pathInfo;
                    }
                    if (sourceInfo) {
                        nodeObject.sourceInfo = sourceInfo;
                    }
                    // TODO: Should be nameKey?
                    if (rawName) {
                        nodeObject.rawName = rawName;
                    }

                    //Ignores any nodes with the parent flow id collapsed
                    var diagramType = shape.parentNode.parentNode.getAttribute("id");
                    if(!diagramType) {
                        diagramType = shape.parentNode.parentNode.id;
                    }

                    if (diagramType !== "IRM_Diagram-Collapse"){
                        nodes.push(nodeObject);
                    }
                }
            }

            SIIUtils.logJSON(nodes,"nodeArray");
            SIIUtils.log("Nodes count: " +nodes.length,true);
            return nodes;
        },

        /**
         * Parses the passed in xml document for link objects.  The results
         * are returned and are also placed in the nodes array;
         * @param {Document} xml The xml document to be parsed for links
         * @returns {Array} A JSON array of the links for the xml document
         */
        parseLinks: function (xml) {
            var edges,
                i,
                path,
                edge,
                bpmnElementId,
                children,
//                t,
//                startX,
//                startY,
//                child,
//                x1,
//                y1,
                element,
                from,
                to,
                name,
                edgeType,
                links,
                link;
            links = [];
            SIIUtils.log("Getting edges");
            resolveNameSpaces(xml);  // Do this here for unit tests
            edges = getElementsByTagNameAgnostic(xml, bpmnDiNamespace, BPMNEDGE_TAGNAME);
            SIIUtils.log("Expecting to create " + edges.length + " links",true);
            for (i = 0; i < edges.length; i++) {
                path = [];
                edge = edges.item(i);
                bpmnElementId = edge.getAttribute(ATTRIBUTE_KEY_BPMN_ELEMENT);
                children = edge.childNodes;

                // Note: We can pass the path as an array of integer values
                // as the "path" attribute to the link.  This should then
                // be bound to the "points" array on the link template.  In my testing
                // I found that it didn't work very well. Routes seemed
                // to have a lot of slanted lines and generally didn't match
                // the output of the editor program, letting GoJS do the
                // routing still seems to be the best fit answer.  I believe the issue
                // has to do with the coordinate system used in the editor and
                // GoJS's calculated coordinate system, they don't seem to completely
                // line up. So for now this code is commented out as it doesn't seem to
                // be needed or useful.
//                for (t = 0; t < children.length; t++) {
//                    child = children[t];
//                    if (!isEmptyNode(child)) {
//                        x1 = child.getAttribute('x');
//                        y1 = child.getAttribute('y');
//                        if (x1 !== null && y1 !== null) {
//                            x1 = parseInt(x1.substring(0, x1.indexOf(".") + 3), 10);
//                            y1 = parseInt(y1.substring(0, y1.indexOf(".") + 3), 10);
//                            path.push(x1);
//                            path.push(y1);
//                        }
//                    }
//                }
//                // The BPMNEdge path seems to include both the start and end points,
//                // which is not required for GoJS as we want it to attach the ports
//                // together.  So we strip off the first and last points in the array.
//                // Note if there are only two points then this results in an empty link
//                if (path.length > 4) {
//                    path = path.slice(2,path.length-2);
//                }
//                else {
//                    path = [];
//                }
                element = getNodeWithId(bpmnElementId, xml);
                from = getSourceRef(element, xml);
                if (!from) {
                    SIIUtils.log("from for edge with id " + bpmnElementId + " is null");
                }
                to = getTargetRef(element, xml);
                if (!to) {
                    SIIUtils.log("to for edge with id " + bpmnElementId + " is null");
                }
                if (from && to) {
                    //name = getElementName(element,false);
                    name = "";
                    edgeType = element.prop(PROPERTY_KEY_TAGNAME);
                    edgeType = removeNameSpace(edgeType);

                    // Check on this
                    if (edgeType === DATA_INPUT_ASSOCIATION_TAG) {
//                        link = {type: edgeType, category: edgeType, path: path, x: startX, y: startY, name: name, from: from, to: to, toPort: "dataInput"};
                        link = {type: edgeType, category: edgeType, name: name, from: from, to: to, fromPort: PORT_VALUE_DATA_OUTPUT, toPort: PORT_VALUE_DATA_INPUT};
                    }
                    else if (edgeType === DATA_OUTPUT_ASSOCIATION_TAG){
//                        link = {type: edgeType, category: edgeType, path: path, x: startX, y: startY, name: name, from: from, to: to, outPort: "dataOutput"};
                        link = {type: edgeType, category: edgeType, name: name, from: from, to: to, fromPort: PORT_VALUE_DATA_OUTPUT, toPort: PORT_VALUE_DATA_INPUT};
                    }
                    else { //sequenceFlow
                        link = {type: edgeType, category: edgeType, name: name, from: from, to: to, fromPort: PORT_VALUE_FLOW_OUTPUT, toPort: PORT_VALUE_FLOW_INPUT};
                    }
                    SIIUtils.logJSON(link,"Adding link");
                    links.push(link);
                }
            }
            SIIUtils.log("Created " + links.length + " links",true);
            return links;
        }
    };
}());